"""
MIT License

Copyright (c) 2021 Overcomer

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import pickle
import logging
import cv2
import numpy as np
from imutils import resize
from sklearn.svm import SVC
from sklearn.preprocessing import LabelEncoder

from new_tools import check_image


class OpenFace(object):
    def __init__(self):
        self.label_encoder = None
        self.embedder_model = None
        self.classifier = None


    def extract_embeddings(self,
                           images_path=list(),
                           vision=False):
        """
        Use OpenFace nn4.small2.v1.t7 network model to extract face feature and output feature vectors.
        
        Args
        ----
        embedding_network_path: OpenFace nn4.small2.v1.t7 network model path.
        
        images_path: Images path.
        
        vision: Show image.
        
        Return
        ------
        embedding_vectors: Face feature vectors.
        """
        
        logging.info("Extracting images...")
        
        total = 0
        embedding_vectors = list()

        for num, image_path in enumerate(images_path, start=1):
            logging.info("Processing {}/{} images...".format(num, len(images_path)))
            
            state, bgr_image = check_image(image_path)
            if state == 0:
                if vision:
                    cv2.imshow("Extracting image...", resize(bgr_image, width=250))
                    cv2.waitKey(10)

                face_blob = cv2.dnn.blobFromImage(bgr_image, 1.0/255, (96, 96), (0, 0, 0), swapRB=True, crop=False)

                self.embedder_model.setInput(face_blob)
                vec = self.embedder_model.forward()

                embedding_vectors.append(vec.flatten())
                total += 1
                
        if vision:
            cv2.destroyAllWindows()
            
        logging.info("Extracted {} count face feature vectors.".format(total))

        return embedding_vectors


    def train_model(self,
                    C=1,
                    kernel="linear",
                    gamma="scale",
                    embedding_vectors=list(),
                    labels=list(),
                    classifier_path=str(),
                    label_encoder_path=str()):
        """
        Train svm classifier by face feature vectors.
        
        Args
        ----
        embedding_vectors: Extract face feature vector by OpenFace nn4.small2.v1.t7 network model.
        
        labels: People name.
        
        classifier_path: Classifier saved path.
        
        label_encoder_path: Label encoder saved path.
        """
        
        logging.info("Encoding labels...")
        label_encoder = LabelEncoder()
        encoded_labels = label_encoder.fit_transform(labels)
        
        logging.info("Training SVM classifier...")
        self.classifier = SVC(C=C, kernel=kernel, gamma=gamma, probability=True)
        self.classifier.fit(embedding_vectors, encoded_labels)

        logging.info("Saving Label encoder...")
        with open(label_encoder_path, "wb") as f: 
            f.write(pickle.dumps(label_encoder))
        if os.path.exists(label_encoder_path): 
            logging.info("Saved label encoder successfully !")
        
        logging.info("Saving classifier...")
        with open(classifier_path, "wb") as f:
            f.write(pickle.dumps(self.classifier))
        if os.path.exists(classifier_path):
            logging.info("Saved classifier successfully !")
            logging.info("Saved classifier to {}.".format(os.path.dirname(classifier_path)))
            
        logging.info("Trained OpenFace classifier finish !")


    def load_model(self,
                   label_encoder_path=str(),
                   classifier_path=str(),
                   embedder_network_path=str()):
        """
        Load OpenFace label encoder、classifier and embedding network model.
        
        Args
        ----
        label_encoder_path: Label encoder saved path.
        
        classifier_path: Classifier saved path.
        
        embedder_network_path: Embedder network model saved path.
        
        Return
        ------
        label_encoder_path: Label encoder saved path.
        
        classifier_path: Classifier saved path.
        
        embedder_network_path: Embedder network model saved path.
        """

        if os.path.exists(label_encoder_path):
            logging.info("Loading label encoder...")
            with open(label_encoder_path, "rb") as lab:
                self.label_encoder = pickle.load(lab)
        
        if os.path.exists(classifier_path):
            logging.info("Loading classifier...")
            with open(classifier_path, "rb") as classifier:
                self.classifier = pickle.load(classifier)
        
        if os.path.exists(embedder_network_path):
            logging.info("Loading OpenFace embedder network model...")
            self.embedder_model = cv2.dnn.readNetFromTorch(embedder_network_path)

        if label_encoder_path == str() and classifier_path == str() and embedder_network_path == str():
            logging.critical("No specify any path !")
            raise ValueError

        logging.info("Loaded OpenFace model finish !")


    def predict(self,
                bgr_image=np.ndarray):
        """
        Use OpenFace nn4.small2.v1.t7 network model predict image.
        
        Args
        ----
        classifier: Classifier saved path.
        
        embedder_network: Embedder network model saved path.
        
        bgr_image: BGR format image.

        Return
        ------
        predict_id: 

        probability:
        """
        
        predict_id = None
        probability = None
        
        state, bgr_image = check_image(bgr_image)
        if state == 0:
            face_blob = cv2.dnn.blobFromImage(bgr_image, 1.0/255, (96, 96), (0, 0, 0), swapRB=True, crop=False)
            self.embedder_model.setInput(face_blob)
            vec = self.embedder_model.forward()

            predicts = self.classifier.predict_proba(vec)[0]
            predict_id = np.argmax(predicts)
            probability = predicts[predict_id]
            logging.debug("Prediction result: {}:{:.6f}".format(predict_id, probability))

        return predict_id, probability


import os
import pickle
import logging
import cv2
import numpy as np
from sklearn.preprocessing import LabelEncoder

from new_tools import check_image


class LBPH(object):
    """
    LBPH class use to show LBPH parameters、trained model、updated model、load model and predict model.

    Attributes:
    -----------
    neighbors: set LBP neighbors point count.

    radius: set LBP circle radius value.

    grid_x: set x axis grid count.

    grid_y: set y axis grid count.

    recognizer: LBPHFaceRecognizer object. 
    """

    def __init__(self,
                 neighbors=8,
                 radius=1,
                 grid_x=8,
                 grid_y=8):
        """
        Init label encoder and LBPH recognizer.

        Args
        ----
        neighbors: LBP sample points.

        radius: LBP radius.

        grid_x: Grid x axis amount.

        grid_y: Grid y axis amount.
        """
        self.label_encoder = None
        self.recognizer = cv2.face.LBPHFaceRecognizer_create(radius=radius, 
                                                             neighbors=neighbors, 
                                                             grid_x=grid_x, 
                                                             grid_y=grid_y)
    @property
    def neighbors(self):
        return self.recognizer.getNeighbors()
    
    @property
    def radius(self):
        return self.recognizer.getRadius()
    
    @property
    def grid_x(self):
        return self.recognizer.getGridX()
    
    @property
    def grid_y(self):
        return self.recognizer.getGridY()
    
    def show_lbph_params(self):
        """
        Show LBPH all parameters.
        """

        logging.info("LBPH Neighbors: {}".format(self.recognizer.getNeighbors()))
        logging.info("LBPH Radius:    {}".format(self.recognizer.getRadius()))
        logging.info("LBPH Grid X:    {}".format(self.recognizer.getGridX()))
        logging.info("LBPH Grid Y:    {}".format(self.recognizer.getGridY()))


    def train_model(self,
                    images=np.array([]),
                    labels=np.array([]),
                    label_encoder_path=str(),
                    model_path=str()):
        """
        train LBPH model and save LBPH model and label encoder object.

        Args:
        -----
        images: Images array.
        
        labels: Labels array.
        
        label_encoder_path: Label encoder path.
        
        model_path: LBPH model path.
        """
        
        logging.info("Encoding labels...")
        label_encoder = LabelEncoder()
        encoded_labels = label_encoder.fit_transform(labels)
        
        with open(label_encoder_path, "wb") as lab_file:
            pickle.dump(label_encoder, lab_file)

        if os.path.exists(label_encoder_path):
            logging.info("Saved label encoder object successfully !")
        else:
            logging.warning("Saved label encoder object failed !")
            
        logging.info("Training model...")
        self.recognizer.train(images, encoded_labels)
        
        self.recognizer.write(model_path)
        if os.path.exists(model_path):
            logging.info("Saved model successfully !")
            logging.info("Saved model to {}.".format(os.path.dirname(model_path)))
        else:
            logging.warning("Saved model failed !")


    def update_model(self, 
                     model_path=str(),
                     images=np.array([]),
                     labels=np.array([])):
        """
        update_model can update the model without retraining.

        Args:
        -----
        model_path: Path of model saves.

        images: Images array.
        
        labels: Labels array.
        """

        logging.info("Updating model...")
        self.recognizer.update(images, labels)

        logging.info("Saving model...")
        self.recognizer.write(model_path)

        logging.info("Update model successfuly !")


    def load_model(self,
                   label_encoder_path=str(),
                   model_path=str()):
        """
        Load LBPH model and label encoder object.
        
        Args:
        -----
        label_encoder_path: Label encoder path.
        
        model_path: Model path.
        """
        if not os.path.exists(label_encoder_path):
            logging.critical("{} path error !".format(label_encoder_path), exc_info=True)
            raise FileNotFoundError
        if not os.path.exists(model_path):
            logging.critical("{} path error !".format(model_path), exc_info=True)
            raise FileNotFoundError

        logging.info("Loading label ecnoder...")
        with open(label_encoder_path, "rb") as lab:
            self.label_encoder = pickle.load(lab)

        logging.info("Loading model...")
        self.recognizer.read(model_path)


    def predict(self,
                gray_image=np.array([])):
        """
        Predict image.
        
        Args
        ----
        gray_image: Gray scale image.
        
        Return
        ------
        predict_id: LBPH model predict ID.
        
        predict_distance: LBPH model predict distance. The smaller the better.
        """

        state, gray_image = check_image(gray_image)
        if state ==0:
            predict_id, predict_distance = self.recognizer.predict(gray_image)
            logging.debug("Prediction result: {}:{:.6f}".format(predict_id, predict_distance))

        return predict_id, predict_distance


"""
MIT License

Copyright (c) 2021 Overcomer

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import logging
import cv2
import dlib
import imutils
from mtcnn import MTCNN
from new_tools import check_image
from ..tools.download import download_models
from ..tools.config import root_dir


class FaceDetection(object):
    """
    FaceDetect class can use four kinds method to detect face. Four methods: Haar Cascade、Dlib Hog、SSD DNN and MTCNN.
    """
    HAAR = 0
    DLIB = 1
    SSD_DNN = 2
    MTCNN = 3


    def load_detector(self, method_ID=int()):
        """
        load_detector method is used to load detector of face detection for reducing each loading time.

        Args:
        -----
        method_ID:
            HAAR: Haar Cascade detector.
            DLIB: Dlib HOG detector.
            SSD_DNN: SSD DNN detector.
            MTCNN: MTCNN detector.
        

        Return:
        -------
        detector: Face detector.

        network: SSD DNN network.
        """
        
        models_id_name = {
                            0: "haarcascade_frontalface_default.xml",
                            2: ["deploy.prototxt", "res10_300x300_ssd_iter_140000.caffemodel"]
                        }
        
        if method_ID == self.HAAR:
            model_name = models_id_name[method_ID]
            model_path = os.path.join(root_dir, model_name)
            
            if not os.path.exists(model_path):
                download_models(model_name=model_name, save_path=root_dir)

            detector = cv2.CascadeClassifier(os.path.join(root_dir, model_name))

            return detector

        elif method_ID == self.DLIB:    
            detector = dlib.get_frontal_face_detector()
            return detector

        elif method_ID == self.SSD_DNN:
            network = None

            for model in models_id_name[method_ID]:
                model_path = os.path.join(root_dir, model)
                
                if not os.path.exists(model_path):
                    download_models(model_name=model, save_path=root_dir)
            
            face_config = os.path.join(root_dir, models_id_name[method_ID][0])
            face_model = os.path.join(root_dir, models_id_name[method_ID][1])
            network = cv2.dnn.readNetFromCaffe(face_config, face_model)

            return network

        elif method_ID == self.MTCNN:
            detector = MTCNN()
            
            return detector


    def haar_detect(self,
                    detector,
                    image=None,  
                    scale_factor=1.025, 
                    min_neighbors=18,
                    min_size=15,
                    max_size=200,
                    vision=False,
                    save_path=None):
        """
        haar_detect method detects face by opencv haar cascade method.

        Args:
        -----
        detector: Haar cascade detector instance.

        image: Input image path or image array.

        scale_factor: Parameter specifying how much the image size is reduced at each image scale.

        min_neighbors: Parameter specifying how many neighbors each candidate rectangle should have to retain it.

        min_size: Minimum possible object size. Objects smaller than that are ignored.

        max_size: Maximum possible object size. Objects larger than that are ignored. If maxSize == minSize model is evaluated on single scale.

        vision: Show image.

        save_path: Save images of detected faces. If vision is False, will doesn't save image.


        Returns:
        --------
        rois: (x, y, w, h)
            x: Face left-top corner x coordinate point.
            y: Face left-top corner y coordinate point.
            w: Width.
            h: Height.

        raw_image: Raw image.

        face_images: Face images.
        """

        rois = list()
        raw_image = None
        face_images = list()

        status, image = check_image(image)
        if status != 0:
            return rois, raw_image, face_images

        raw_image = image.copy()
        draw_image = image.copy()

        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    
        faces = detector.detectMultiScale(gray,
                                          scaleFactor=scale_factor,
                                          minNeighbors=min_neighbors,
                                          minSize=(min_size, min_size),
                                          maxSize=(max_size, max_size))
    
        logging.debug("detection.haar_detect.faces count: {}".format(len(faces)))

        if len(faces) > 0:
            for roi in faces:    
                (x, y, w, h) = roi
                rois.append(roi)
                face_images.append(image.copy()[y:y+h, x:x+w])

            if vision:
                for num, (x,y,w,h) in enumerate(rois, start=1):
                    cv2.rectangle(draw_image, (x, y), (x+w, y+h), (0, 255, 0), 3)
                    cv2.putText(draw_image, str(num), (x, y-5), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 3)
        else:
            logging.info("Haar Cascade doesn't detected the faces !")

        if vision:
            text = "Haar Detected Faces: {}".format(len(face_images))
            cv2.putText(draw_image, text, (0, 30), cv2.FONT_HERSHEY_SIMPLEX, 1.25, (0, 0, 255), 3)
            cv2.imshow('Haar Cascade Face Detection', imutils.resize(draw_image, width=640))
            cv2.waitKey(0)
            cv2.destroyAllWindows()

        if save_path != None:
            cv2.imwrite(save_path, draw_image)

            if not os.path.exists(save_path):
                logging.error("{} saved failed !".format(save_path))
                raise FileNotFoundError

        logging.info("Haar Cascade detected faces: {}".format(len(face_images)))
        return (rois, raw_image, face_images)
    
    
    def dlib_detect(self,
                    detector,
                    image=None,
                    unsample=0,
                    conf_threshold=0.15,
                    vision=False,
                    save_path=None):
        """
        dlib_detect method detects face by HOG and SVM method.

        Args:
        -----
        detector: Dlib face detector instance.

        image: Input image path or image array.
        
        unsample: Number of times to upsample an image before applying face detection. Default 0 is fastest.

        conf_threshold: conf_threshold value is used to judge the face detection true or false.

        vision: Show image.

        save_path: Save images of detected faces. If vision is False, will doesn't save image. 


        Returns:
        --------
        rois: (x, y, w, h)
            x: Face left-top corner x coordinate point.
            y: Face left-top corner y coordinate point.
            w: width.
            h: Height.

        raw_image: Raw image.

        face_images: Face images.
        """

        rois = list()
        raw_image = None
        face_images = list()

        status, image = check_image(image)
        if status != 0:
            return rois, raw_image, face_images

        raw_image = image.copy()
        draw_image = image.copy()

        faces, scores, idxs = detector.run(image, unsample, conf_threshold)
        logging.debug("detection.dlib_detect.faces count: {}".format(len(faces)))
        logging.debug("detection.dlib_detect.scores count: {}".format(len(scores)))
        logging.debug("detection.dlib_detect.idxs count: {}".format(len(idxs)))
    
        if len(faces) > 0:
            for face in faces:
                roi = self.__rect_to_roi(face)
                rois.append(roi)
                x, y, w, h = roi
                face_images.append(image.copy()[y:h, x:w])

            if vision:
                for num, (x, y, w, h) in enumerate(rois, start=1):
                    cv2.rectangle(draw_image, (x, y), (w, h), (0, 255, 0), 3)
                    cv2.putText(draw_image, str(num), (x, y-5), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 3)
        else:
            logging.info("Dlib HoG doesn't detected the faces !")

        if vision:
            text = "Dlib HoG Detected Faces: {}".format(len(face_images))
            cv2.putText(draw_image, text, (0, 30), cv2.FONT_HERSHEY_SIMPLEX, 1.25, (0, 0, 255), 3)
            cv2.imshow("Dlib HoG Face Detection", imutils.resize(draw_image, width=640))
            cv2.waitKey(0)
            cv2.destroyAllWindows()

        if save_path != None:
            cv2.imwrite(save_path, draw_image)

            if not os.path.exists(save_path):
                logging.error("{} saved failed !".format(save_path))
                raise FileNotFoundError

        logging.info("Dlib HOG+SVM detected faces: {}".format(len(face_images)))
        return (rois, raw_image, face_images)


    def ssd_dnn_detect(self,
                       network,
                       image=None,
                       conf_threshold=0.7,
                       vision=False,
                       save_path=None):
        """
        ssd_dnn_detect method detects face by tensorflow or caffe ssd dnn framework.

        Args:
        -----
        network: Caffe or TensorFlow framework instance.

        image: Input image path or image array.

        conf_threshold: conf_threshold value is used to judge the face detection true or false.

        vision: Show image.

        save_path: Save images of detected faces. If vision is False, will doesn't save image. 


        Returns:
        --------
        rois:
            (x, y, w, h)
                x: Face left-top corner x coordinate point.
                y: Face left-top corner y coordinate point.
                w: Width.
                h: Height.

        raw_image: Raw image.

        face_images: Face images.
        """

        rois = list()
        raw_image = None
        face_images = list()

        status, image = check_image(image)
        if status != 0:
            return rois, raw_image, face_images
            
        raw_image = image.copy()
        draw_image = image.copy()

        frameHeight = image.shape[0]
        frameWidth = image.shape[1]

        blob = cv2.dnn.blobFromImage(image, 1.0, (300, 300), [104, 117, 123], swapRB=False, crop=False)

        network.setInput(blob)
        detections = network.forward()
        for i in range(detections.shape[2]):
            confidence = detections[0, 0, i, 2]
            if confidence > conf_threshold:
                x = int(detections[0, 0, i, 3] * frameWidth)
                y = int(detections[0, 0, i, 4] * frameHeight)
                w = int(detections[0, 0, i, 5] * frameWidth)
                h = int(detections[0, 0, i, 6] * frameHeight)
                
                rois.append((x, y, w, h))
                face_images.append(image.copy()[y:h, x:w])
        logging.debug("detection.ssd_dnn_detect.face_images count: {}".format(len(face_images)))
                
        if vision:
            if len(face_images) > 0:
                for num, (x, y, w, h) in enumerate(rois, start=1):
                    cv2.rectangle(draw_image, (x, y), (w, h), (0, 255, 0), 3)
                    cv2.putText(draw_image, str(num), (x, y-5), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 3)
            else:
                logging.info("SSD DNN doesn't detected the faces !")

            text = "SSD DNN Detected Faces: {}".format(len(face_images))
            cv2.putText(draw_image, text, (0, 30), cv2.FONT_HERSHEY_SIMPLEX, 1.25, (0, 0, 255), 3)
            cv2.imshow("SSD DNN Face Detection", imutils.resize(draw_image, width=640))
            cv2.waitKey(0)
            cv2.destroyAllWindows()

        if save_path != None:
            cv2.imwrite(save_path, draw_image)
            
            if not os.path.exists(save_path):
                logging.error("{} saved failed !".format(save_path))
                raise FileNotFoundError

        logging.info("SSD DNN detected faces: {}".format(len(face_images)))
        return (rois, raw_image, face_images)


    def mtcnn_detect(self,
                     detector,
                     image=None,
                     min_face_size=20,
                     conf_threshold=0.75,
                     vision=False,
                     save_path=None):
        """
        mtcnn_detect method detects face by multi-task cascaded convolutional networks method.

        Args:
        -----
        detector: MTCNN object.

        image: Input image path or image array.

        min_face_size: Face detection minimum size. Default 20.

        conf_threshold: conf_threshold value is used to judge the face detection true or false.
        
        vision: Show image.

        save_path: Save images of detected faces. If vision is False, will doesn't save image.
        

        Returns:
        --------
        rois: (x, y, w, h)
            x: Face left-top corner x coordinate point.
            y: Face left-top corner y coordinate point.
            w: Width.
            h: Height.

        raw_image: Raw image.

        face_images: Face images.
        """
        
        rois = list()
        raw_image = None
        face_images = list()

        status, image = check_image(image)
        if status != 0:
            return rois, raw_image, face_images
            
        raw_image = image.copy()
        draw_image = image.copy()

        rgb_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        detector.min_face_size = min_face_size
        result = detector.detect_faces(rgb_image)
        logging.debug("detection.mtcnn_detect.result count: {}".format(len(result)))

        if len(result) > 0:
            for persorn in result:
                if persorn["confidence"] >= conf_threshold:
                    x, y, w, h = persorn["box"]
                    rois.append((x, y, w, h))
                    face_images.append(raw_image.copy()[y:y+h, x:x+w])

            if vision:
                for num, (x,y,w,h) in enumerate(rois, start=1):
                    cv2.rectangle(draw_image, (x, y), (x+w, y+h), (0, 255, 0), 3)
                    cv2.putText(draw_image, str(num), (x, y-5), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 0, 255), 3)
        else:
            logging.info("MTCNN doesn't detected the faces ! ")

        if vision:
            text = "MTCNN Detected Faces: {}".format(len(face_images))
            cv2.putText(draw_image, text, (0, 30), cv2.FONT_HERSHEY_SIMPLEX, 1.25, (0, 0, 255), 3)
            cv2.imshow("MTCNN Face Detection", imutils.resize(draw_image, width=640))
            cv2.waitKey(0)
            cv2.destroyAllWindows()

        if save_path != None:
            cv2.imwrite(save_path, draw_image)
            
            if not os.path.exists(save_path):
                logging.error("{} saved failed !".format(save_path))
                raise FileNotFoundError

        logging.info("MTCNN detected faces: {}".format(len(face_images)))
        return (rois, raw_image, face_images)


    def __rect_to_roi(self, rect):
        """
        __rect_to_roi method is used to get roi values from dlib.rectangle class.

        Args:
        -----
        rect: Face roi.
        
        
        Return:
        -------
        rois:
            (x, y, w, h)
                x: Face left-top corner x coordinate point.
                y: Face left-top corner y coordinate point.
                w: Width.
                h: Height.
        """

        roi = (rect.left(), rect.top(), rect.right(), rect.bottom())
        logging.debug("detection.__rect_to_roi.roi: {}".format(roi))

        return roi


"""
MIT License

Copyright (c) 2021 Overcomer

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import pickle
import logging
import numpy
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.utils import plot_model
from tensorflow.keras.models import load_model
from tensorflow.keras.models import Sequential
from tensorflow.keras import layers, regularizers
from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint

from .openface import OpenFace
from new_tools import check_image


class LBPCNN(OpenFace):
    def __init__(self):
        """
        Init label_encoder and model.
        """

        self.label_encoder = None
        self.model = None


    def build_LBPCNN_model(self, name="LBPCNN", classes=18):
        """
        LBPCNN architecture.

        Args
        ----
        name: Specify CNN model name.

        classes: Specify classes amount.
        """
        model = Sequential(name=name)

        model.add(layers.Conv2D(filters=40, kernel_size=(3, 3), padding="same", input_shape=(256, 256, 1), activation="relu", name="cnn1_1"))
        model.add(layers.Conv2D(filters=40, kernel_size=(3, 3), padding="same", activation="relu", name="cnn1_2"))
        model.add(layers.BatchNormalization(name="bn1"))
        model.add(layers.MaxPooling2D(pool_size=(2, 2), strides=2, name="mp1"))

        model.add(layers.Conv2D(filters=64, kernel_size=(3, 3), padding="same", activation="relu", name="cnn2_1"))
        model.add(layers.Conv2D(filters=64, kernel_size=(3, 3), padding="same", activation="relu", name="cnn2_2"))
        model.add(layers.BatchNormalization(name="bn2"))
        model.add(layers.MaxPooling2D(pool_size=(2, 2), strides=2, name="mp2"))

        model.add(layers.Conv2D(filters=128, kernel_size=(3, 3), strides=1, padding="same", activation="relu", name="cnn3_1"))
        model.add(layers.Conv2D(filters=128, kernel_size=(3, 3), strides=1, padding="same", activation="relu", name="cnn3_2"))
        model.add(layers.BatchNormalization(name="bn3"))
        model.add(layers.MaxPooling2D(pool_size=(2, 2), strides=2, name="mp3"))

        model.add(layers.Conv2D(filters=64, kernel_size=(3, 3), padding="same", activation="relu", name="cnn4_1"))
        model.add(layers.Conv2D(filters=64, kernel_size=(3, 3), padding="same", activation="relu", name="cnn4_2"))
        model.add(layers.BatchNormalization(name="bn4"))
        model.add(layers.MaxPooling2D(pool_size=(2, 2), strides=2, name="mp4"))

        model.add(layers.Conv2D(filters=40, kernel_size=(3, 3), padding="same", activation="relu", name="cnn5_1"))
        model.add(layers.Conv2D(filters=40, kernel_size=(3, 3), padding="same", activation="relu", name="cnn5_2"))
        model.add(layers.BatchNormalization(name="bn5"))
        model.add(layers.MaxPooling2D(pool_size=(2, 2), strides=2, name="mp5"))

        model.add(layers.Flatten(name="flatten"))

        model.add(layers.Dense(256, activation="relu", name="fc1"))
        model.add(layers.Dropout(0.25, name="dp1"))
        model.add(layers.BatchNormalization(name="bn6"))
        
        model.add(layers.Dense(128, kernel_regularizer=regularizers.L1L2(0.0001), bias_regularizer=regularizers.L1L2(0.0001), activation="relu", name="fc2"))
        model.add(layers.Dropout(0.25, name="dp2"))
        model.add(layers.BatchNormalization(name="bn7"))

        model.add(layers.Dense(28, activation="softmax", name="softmax"))

        model.compile(loss="sparse_categorical_crossentropy",
                    optimizer=Adam(learning_rate=2.5e-4),
                    metrics=["accuracy"])

        model.summary()

        self.model = model
    
    
    def load_model(self,
                   label_encoder_path=str(),
                   model_path=str()):
        """
        Load LBPCNN model from SaveModel class. You need to specify model diretory path.

        Args
        ----
        label_encoder_path: Specify label encoder path.

        model_path: Specify LBPCNN model path.
        """
        
        logging.info("Loading LBPCNN model...")
        if not os.path.exists(label_encoder_path):
            logging.critical("'{}' error ! Loaded LBPCNN label encoder failed.".format(label_encoder_path))
            raise FileNotFoundError

        if not os.path.exists(model_path):
            logging.critical("'{}' error ! Loaded LBPCNN model failed.".format(model_path))
            raise FileNotFoundError
            
        with open(label_encoder_path, "rb") as lab:
            self.label_encoder = pickle.load(lab)

        self.model = load_model(model_path)
        
        logging.info("Loaded LBPCNN model successfully !")

    
    def train_model(self,
                    images,
                    labels,
                    validation_data=tuple(),
                    model=Sequential(),
                    epochs=200,
                    batch_size=128,
                    verbose=1,
                    time_record=str(),
                    save_path=str()):
        """
        Train LBPCNN model.
        
        Aegs
        ----
        images: Training images array.
        
        labels: Training labels.
        
        validation_data: Input validation images array and validation labels.
        
        model: Sequential model instance.
        
        epochs: Training model count.
        
        batch_size: Images amount of once training model. Power of 2 like 2,4,8,...,128.
        
        verbose: Training model real time detail information.
        
        time_record: Build training model directory prefixes.
        
        save_path: Save model root directory.
        
        Return
        ------
        train_history: LBPCNN model training history.
        """
        
        logging.info("Training LBPCNN model...")
        
        root_path = os.path.join(save_path, "{}".format(time_record))
        logging.info("Building LBPCNN model saved path to {}...".format(root_path))
        if not os.path.join(root_path):
            os.makedirs(root_path)
        
        plot_image = os.path.join(root_path, "{}_LBPCNN_model_architecture.png".format(time_record))
        logging.info("Saving LBPCNN model architecture image to {}...".format(plot_image))
        plot_model(model,
                   to_file=plot_image,
                   show_shapes=True,
                   show_layer_names=True)
        
        log_dir = os.path.join(root_path, "logs\\fit\\{}".format(time_record))
        logging.info("Building tensorboard log directory to {}...".format(log_dir))
        os.makedirs(log_dir)
        
        tensorboard_callback = TensorBoard(log_dir=log_dir,
                                           histogram_freq=1,
                                           write_graph=True,
                                           write_images=True,
                                           embeddings_freq=False)

        checkpoint_dir = os.path.join(root_path, "{}_best_checkpoint".format(time_record))
        logging.info("Building checkpoint directory to {}...".format(checkpoint_dir))
        os.makedirs(checkpoint_dir)
        
        checkpoint_filepath = os.path.join(checkpoint_dir, "lbpcnn_{epoch:04d}.ckpt")
        checkpoint_callback = ModelCheckpoint(filepath=checkpoint_filepath,
                                              monitor='val_loss',
                                              mode='min',
                                              verbose=0,
                                              save_best_only=True)
        train_history = model.fit(ｘ=images,
                                  ｙ=labels,
                                  validation_data=validation_data,
                                  epochs=epochs,
                                  batch_size=batch_size,
                                  verbose=verbose,
                                  callbacks=[tensorboard_callback, checkpoint_callback])
        logging.info("Saved LBPCNN model to {}.".format(root_path))
        logging.info("Finished LBPCNN model training !")

        return train_history
        
    
    def predict(self,
                image):
        """
        Predict image.

        Args
        ----
        image: LBP image format.

        Return
        ------
        pred_id: Prediction index.

        pred_proba: Prediction probability.
        """
        
        pred_id = None
        pred_proba = None
        
        state, image = check_image(image)
        if state == 0:
            predition = self.model.predict(image)[0]
            logging.debug("lbpcnn.LBPCNN.predict.prediction: {}".format(predition))
            logging.debug("lbpcnn.LBPCNN.predict.prediction shape: {}".format(predition.shape))
            pred_id = numpy.argmax(predition)
            pred_proba = predition[pred_id]
            
            logging.debug("Prediction result: {}:{:.6f}".format(pred_id, pred_proba))
            
        return pred_id, pred_proba


"""
MIT License

Copyright (c) 2021 Overcomer

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import logging
import cv2
import dlib
from new_tools import check_image
from ..tools.config import root_dir
from ..tools.download import download_models

class FaceLandmark(object):
    """
    FaceLandmark class use two methods to landmark face.
    """

    @classmethod
    def load_shape_predictor(cls,
                             model_name="shape_predictor_5_face_landmarks.dat"):
        """
        load_shape_predictor method is used to load dlib shape predictor.

        Args
        ----
        model_name: shape_predictor_5_face_landmarks.dat or shape_predictor_68_face_landmarks.dat file names.


        Return
        ------
        shape_predictor: shape_predictor instance.
        """

        shape_predictor_path = os.path.join(root_dir, model_name)

        if not os.path.exists(shape_predictor_path):
            download_models(model_name, save_path=root_dir)

        shape_predictor = dlib.shape_predictor(shape_predictor_path)
        
        return shape_predictor


    @classmethod
    def dlib_5_points(cls,
                      image=None,
                      shape_predictor=object(),
                      vision=False,
                      save_path=None):
        """
        dlib_5_points method is use face five points of dlib to mark left eye, right eye, nose of face.

        Args:
        -----
        image: Input image path or image array.

        shape_predictor: shape_predictor instance.

        vision: Show image.

        save_path: Save images of detected faces. If vision is False, will doesn't save image.


        Returns:
        --------
        five_points:
            lefteye_leftcorner: left eye corner coordinate of left eye.
            lefteye_rightcorner: Right eye corner coordinate of left eye.
            righteye_rightcorner: Right eye corner coordinate of right eye.
            righteye_leftcorner: Left eye corner coordinate of right eye.
            nose: Nose coordinate.
        """

        five_points = dict()
        five_points_dict = dict()

        status, raw_image = check_image(image)
        if status != 0:
            return five_points

        detector = dlib.get_frontal_face_detector()

        detect_faces = detector(raw_image, 2)
        
        if len(detect_faces) > 0:
            for num, roi in enumerate(detect_faces):
                five_points = dict()

                shape_face = shape_predictor(raw_image, roi)
                lefteye_leftcorner, lefteye_rightcorner, righteye_rightcorner, righteye_leftcorner, nose = shape_face.parts()
                five_points["lefteye_leftcorner"] = (lefteye_leftcorner.x, lefteye_leftcorner.y)
                five_points["lefteye_rightcorner"] = (lefteye_rightcorner.x, lefteye_rightcorner.y)
                five_points["righteye_rightcorner"] = (righteye_rightcorner.x, righteye_rightcorner.y)
                five_points["righteye_leftcorner"] = (righteye_leftcorner.x, righteye_leftcorner.y)
                five_points["nose"] = (nose.x, nose.y)
                five_points_dict[num] = five_points

            if vision:
                for item in five_points_dict:
                    person = five_points_dict[item]
                    for point in person:
                        coordinate = person[point]
                        cv2.circle(raw_image, coordinate, 2, (0, 255, 0), 2)

                cv2.imshow("Face Landmark", raw_image)
                cv2.waitKey(0)
                cv2.destroyAllWindows()

            if save_path != None:
                cv2.imwrite(save_path, raw_image)
                
                if not os.path.exists(save_path):
                    logging.error("{} saved failed !".format(save_path))
                    raise FileNotFoundError
        else:
            logging.info("Dlib doesn't detect the face !")
        
        return five_points_dict


    @classmethod
    def dlib_68_points(cls,
                       image=None,
                       shape_predictor=object(),
                       get_five_points=False,
                       vision=False,
                       save_path=None):
        """
        dlib_68_points method is use face sixty-eight points of dlib to mark sixty-eight of face.
        
        Args:
        -----
        image: Input image path or image array.

        shape_predictor: shape_predictor instance.

        get_five_points: Control only get five points of face from sixty-eight points of face.

        vision: Show image.

        save_path: Save images of detected faces. If vision is False, will doesn't save image.

        
        Returns:
        --------
        five_points_dict:
            lefteye_leftcorner: left eye corner coordinate of left eye.
            lefteye_rightcorner: Right eye corner coordinate of left eye.
            righteye_rightcorner: Right eye corner coordinate of right eye.
            righteye_leftcorner: Left eye corner coordinate of right eye.
            nose: Nose coordinate.

        sixty_eight_points_dict: Sixty eight points of face.
        """

        five_points = dict()
        sixty_eight_points = dict()
        five_points_dict = dict()
        sixty_eight_points_dict = dict()


        status, raw_image = check_image(image)
        if status != 0:
            return sixty_eight_points
            
        detector = dlib.get_frontal_face_detector()

        detect_faces = detector(raw_image, 2)

        if len(detect_faces) > 0:
            for num, roi in enumerate(detect_faces):
                shape_face = shape_predictor(raw_image, roi)

                if get_five_points:
                    five_points = dict()
                    five_points["lefteye_leftcorner"] = (shape_face.part(36).x, shape_face.part(36).y)
                    five_points["lefteye_rightcorner"] = (shape_face.part(39).x, shape_face.part(39).y)
                    five_points["righteye_rightcorner"] = (shape_face.part(45).x, shape_face.part(45).y)
                    five_points["righteye_leftcorner"] = (shape_face.part(42).x, shape_face.part(42).y)
                    five_points["nose"] = (shape_face.part(30).x, shape_face.part(30).y)
                    five_points_dict[num] = five_points
                else:
                    sixty_eight_points = dict()
                    for i in range(0, 68):
                        sixty_eight_points[i] = (shape_face.part(i).x, shape_face.part(i).y)

                    sixty_eight_points_dict[num] = sixty_eight_points
            
            if vision:
                if get_five_points:
                    for item in five_points_dict:
                        person = five_points_dict[item]
                        for point in person:
                            coordinate = person[point]
                            cv2.circle(raw_image, coordinate, 2, (0, 255, 0), 2)
                else:       
                    for item in sixty_eight_points_dict:
                        person = sixty_eight_points_dict[item]
                        for point in person:
                            coordinate = person[point]
                            cv2.circle(raw_image, coordinate, 2, (0, 255, 0), 2)

                cv2.imshow("Face Landmark", raw_image)
                cv2.waitKey(0)
                cv2.destroyAllWindows()
            
            if save_path != None:
                cv2.imwrite(save_path, raw_image)
                
                if not os.path.exists(save_path):
                    logging.error("{} saved failed !".format(save_path))
                    raise FileNotFoundError
        else:
            logging.info("Dlib doesn't detect the face !")

        if get_five_points:
            return five_points_dict
        else:
            return sixty_eight_points_dict

    
    @classmethod
    def __calc_center_point(cls, x1=int(), y1=int(), x2=int(), y2=int()):
        """
        __calc_center_point method is used to calculate center coordinate of two point.

        Args:
        -----
        x1: x coordinate of x1.

        y1: y coordinate of x1.

        x2: x coordinate of x2.

        y2: y coordinate of x2.

        Returns:
        --------
        (x, y): Center coordinate.
        """

        (x, y) = (int((x1 + x2) / 2), int((y1 + y2) / 2))
        return (x, y)

    
    @classmethod
    def fivepoints_to_threepoints(cls, five_point=dict()):
        """
        fivepoints_to_threepoints method used to transfer 5 points to 3 points.

        Args:
        -----
        5 points: 
            lefteye_leftcorner: left eye corner coordinate of left eye.
            lefteye_rightcorner: Right eye corner coordinate of left eye.
            righteye_rightcorner: Right eye corner coordinate of right eye.
            righteye_leftcorner: Left eye corner coordinate of right eye.
            nose: Nose coordinate.


        Return:
        -------
        three_points:
            left_eye: left eye center coordinate.
            right_eye: Right eye center coordinate.
            nose: Nose coordinate.
        """
        
        three_points = dict()
        if len(five_point) < 5:
            logging.error("five_point variable element small than 5 !")
            raise ValueError
        lefteye_leftcorner_x1, lefteye_leftcorner_y1 = five_point["lefteye_leftcorner"]
        lefteye_rightcorner_x2, lefteye_rightcorner_y2 = five_point["lefteye_rightcorner"]

        three_points["left_eye"] = cls.__calc_center_point(lefteye_leftcorner_x1, 
                                                          lefteye_leftcorner_y1, 
                                                          lefteye_rightcorner_x2, 
                                                          lefteye_rightcorner_y2)

        righteye_leftcorner_x1, righteye_leftcorner_y1 = five_point["righteye_leftcorner"]
        righteye_rightcorner_x2, righteye_rightcorner_y2 = five_point["righteye_rightcorner"]

        three_points["right_eye"] = cls.__calc_center_point(righteye_leftcorner_x1,
                                                           righteye_leftcorner_y1, 
                                                           righteye_rightcorner_x2,
                                                           righteye_rightcorner_y2)
        three_points["nose"] = five_point["nose"]

        return three_points


"""
MIT License

Copyright (c) 2021 Overcomer

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import math
import logging
import cv2
import dlib
from mtcnn import MTCNN
import imutils
from imutils.face_utils import FaceAligner
from new_tools import check_image
from ..tools.config import root_dir
from ..tools.download import download_models


class FaceAlignment(object):
    """
    FaceAlignment class can use two kind method to alignment face. Two method: MTCNN and Dlib.
    """
    MTCNN = 0
    DLIB = 1


    def __compute_center_point(self, left_eye=tuple(), right_eye=tuple()):
        """
        Compute center point of two eyes.

        Args
        ----
        left_eye: x and y axis of left eye.

        right_eye: x and y axis of right eye.

        Return
        -------
        center_point: Center point axis of two eyes.
        """
        
        x = int((left_eye[0] + right_eye[0]) / 2)
        y = int((left_eye[1] + right_eye[1]) / 2)
        center_point = (x, y)

        return center_point


    def __compute_degree(self, left_eye=tuple(), right_eye=tuple()):
        """
        __compute_degree method is used to compute slope of two eyes, and transform unit from radian to degree.

        Args:
        -----
        left_eye: Left eye coordinate point x and y.

        right_eye: Right eye coordinate point x and y.


        Return:
        --------
        rotation_degree: Rotational degree.
        """

        if len(left_eye) != 2 or len(right_eye) != 2:
            logging.debug("__compute_degree.left_eye: {}".format(left_eye))
            logging.debug("__compute_degree.right_eye: {}".format(right_eye))
            logging.error("left eye or right eye coordinate less than two !", exc_info=True)
            raise ValueError
        lefteye_x, lefteye_y = left_eye
        righteye_x, righteye_y = right_eye

        rotation_degree = math.degrees(math.atan2(righteye_y - lefteye_y, righteye_x - lefteye_x))
        logging.debug("FaceAlignment.__compute_degree.rotation_degree: {:.2f}°".format(rotation_degree))

        return rotation_degree


    def __area_expand(self, roi=tuple(), width_scale_factor=0.1):
        """
        __area_expand method is used to expand face area.

        Args:
        -----
        roi: (x, y, w, h)
            x: Face left-top corner x coordinate point.
            y: Face left-top corner y coordinate point.
            w: Width.
            h: Height.


        Return:
        --------
        (nx, ny, nw, nh):
            nx: New face left-top corner x coordinate point.
            ny: New face left-top corner y coordinate point.
            nw: New width.
            nh: New height.
        """

        if len(roi) != 4: 
            logging.debug("FaceAlignment.__area_expand.roi: {}".format(roi))
            logging.error("roi values less than four.", exc_info=True)
            raise ValueError
        
        x, y, w, h = roi
        width_scale_factor = 0.1
        nx = int(x - (width_scale_factor * w))
        ny = y
        nw = int((1 + width_scale_factor * 2) * (w))
        nh = h

        if nx < 0:
            nx = 0 
        if ny < 0:
            ny = 0
        
        logging.debug("FaceAlignment.__area_expand.roi: {}".format(roi))
        logging.debug("FaceAlignment.__area_expand new roi: {}".format((nx, ny, nw, nh)))
        return (nx, ny, nw, nh)


    def __dlib_rect_to_roi(self, rect):
            """
            __dlib_rect_to_roi method is used to get roi values from dlib.rectangle class.

            Args:
            -----
                rect: Face roi


            Return:
            -------
            roi:
                (x, y, w, h)
                    x: Face left-top corner x coordinate point.
                    y: Face left-top corner y coordinate point.
                    w: Width.
                    h: Height.                     
            """

            roi = (rect.left(), rect.top(), rect.right(), rect.bottom())
            logging.debug("__dlib_rect_to_roi.roi: {}".format(roi))

            return roi

    
    def load_detector(self, method_ID=int()):
        """
        load_detector method is used to load all method detector for reducing loading time.

        Args:
        -----
        method_ID:
            FaceAlignment.MTCNN: Load mtcnn detector.
            FaceAlignment.DLIB:  Load dlib face detector、shape predictor and face aligner of imutils.
        

        Return:
        -------
        detector: Face detector.

        face_aligner: imutils face aligner, only return when use FaceAlignment.DLIB.
        """
        
        if type(method_ID) != int: 
            logging.debug("load_detector.method_ID: {}".format(method_ID))
            logging.error("method_ID type isn't int.", exc_info=True)
            raise TypeError
        elif method_ID == FaceAlignment.MTCNN:
            logging.info("Loading mtcnn detector...")
            detector = MTCNN()
            return detector
        elif method_ID == FaceAlignment.DLIB:
            logging.info("Loading dlib detector...")

            model_name = "shape_predictor_68_face_landmarks.dat"
            shape_landmark_file_path = os.path.join(root_dir, model_name)
            if not os.path.exists(shape_landmark_file_path):
                download_models(model_name, root_dir)

            face_size = 256

            detector = dlib.get_frontal_face_detector()

            if not os.path.exists(shape_landmark_file_path): 
                logging.error("{} path error !".format(shape_landmark_file_path), exc_info=True)
                raise FileNotFoundError
            shape_predictor = dlib.shape_predictor(shape_landmark_file_path)

            face_aligner = FaceAligner(shape_predictor, desiredFaceWidth=face_size)

            return detector, face_aligner
        else:
            logging.warning("Not match any method !")


    def mtcnn_alignment(self,
                        detector,
                        image,
                        conf_threshold=0.75,
                        vision=False,
                        vision_millisecs=100,
                        save_dir=None,
                        face_size=256):
        """
        mtcnn_alignment method is used to alignment face by multi-task cascaded convolutional networks method.
        
        ※Notice: MTCNN need to RGB image, if you use cv2.imread() to read image, you need swap R and B channel.

        Args:
        -----
        detector: Input MTCNN instance.
        
        image: Image can input image array or image path.

        conf_threshold: conf_threshold value is used to judge the face detection true or false.

        vision: Show face alignment image.

        vision_millisecs: Show image seconds. 
        
        save_dir: Saving path of face alignment images.

        face_size: Saved face size.


        Return:
        --------
        rois:
            (x, y, w, h)
                x: Face left-top corner x coordinate point.
                y: Face left-top corner y coordinate point.
                w: Width.
                h: Height.

        raw_image: Raw image.

        face_images: Image of face alignment.
        """
        
        rois = list()
        raw_image = None
        face_images = list()

        status, image = check_image(image)
        if status !=0:
            return rois, raw_image, face_images

        logging.debug("alignment.FaceAlignment.mtcnn.alignment.image shape: {}".format(image.shape))
        raw_image = image.copy()

        rgb_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        logging.debug("Detecting face...")
        result = detector.detect_faces(rgb_image)
        logging.debug("mtcnn_alignment.result: {}".format(result))

        logging.debug("Aligning face...")
        if len(result) > 0:
            for num, people in enumerate(result, start=1):
                if people["confidence"] >= conf_threshold:
                    face_point = people["keypoints"]
                    
                    (img_h, img_w) = rgb_image.shape[:2]

                    lefteye = face_point["left_eye"]
                    righteye = face_point["right_eye"]

                    rotation_point = self.__compute_center_point(lefteye, righteye)

                    rotation_degree = self.__compute_degree(lefteye, righteye)

                    M = cv2.getRotationMatrix2D(rotation_point, rotation_degree, scale=1.0)
                    rgb_rotated = cv2.warpAffine(rgb_image, M, (img_w, img_h), flags=cv2.INTER_CUBIC)
                    
                    x, y, w, h = roi = people['box']
                    rois.append(roi)

                    nx, ny, nw, nh = self.__area_expand(roi)
                    align_face_image = cv2.cvtColor(rgb_rotated[ny:ny+nh, nx:nx+nw], cv2.COLOR_RGB2BGR)

                    align_face_image = cv2.resize(align_face_image, (face_size, face_size))
                    face_images.append(align_face_image)

                    if vision:
                        cv2.imshow("MTCNN Raw Image...", imutils.resize(raw_image, width=640))
                        raw_face_image = image[y:y+h, x:x+w]
                        cv2.imshow("Raw Face Image...", imutils.resize(raw_face_image, width=250))
                        cv2.imshow("MTCNN Align faces...", imutils.resize(align_face_image, width=250))
                        cv2.waitKey(vision_millisecs)

                    if save_dir != None:
                        if not os.path.exists(save_dir):
                            os.makedirs(save_dir)
                            logging.info("Builed {} directory successfully !") if os.path.exists(save_dir) else logging.warning("Builed {} directory failed !")

                        image_path = os.path.join(save_dir, "{}.jpg".format(str(num).zfill(6)))
                        cv2.imwrite(image_path, align_face_image, [cv2.IMWRITE_JPEG_QUALITY, 100])
                        
                        if os.path.exists(image_path):
                            logging.info("Saved '{}' successfully !".format(image_path))
                        else:
                            logging.warning("Saved '{}' failed !".format(image_path))
            cv2.destroyAllWindows()
        else:
            logging.debug("MTCNN doesn't detect the face !")

        return (rois, raw_image, face_images)


    def dlib_alignment(self,
                       detector,
                       face_aligner,
                       image,
                       vision=False,
                       vision_millisecs=100,
                       save_dir=None,
                       face_size=256):
        """
        dlib_alignment method is used to alignment face by 5 or 68 point landmark method.
        It can use shape 5 or 68 point landmark file.

        Args:
        -----
        detector: Input dlib face detector instance.

        face_aligner: FaceAligner instance of imutils.

        image: Image array or image path.

        vision: Show face alignment image.

        vision_millisecs: Show image seconds.
        
        save_dir: Saving path of face alignment image.

        face_size: Saved face size.


        Return:
        --------
        rois:
            (x, y, w, h)
                x: Face left-top corner x coordinate point.
                y: Face left-top corner y coordinate point.
                w: Width.
                h: Height.

        raw_image: Raw image.

        face_image: Image of face alignment.
        """
        
        rois = list()
        raw_image = None
        face_images = list()

        status, image = check_image(image)
        if status !=0:
            return rois, raw_image, face_images

        raw_image = image.copy()

        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        logging.debug("Detecting face...")
        rects = detector(gray, 2)
        logging.debug("alignment.dlib_alignment.rects: {}".format(rects))

        logging.debug("Aligning face image...")
        if len(rects) > 0:
            for rect in rects:
                roi = self.__dlib_rect_to_roi(rect)
                face_image = face_aligner.align(image, gray, rect)

                rois.append(roi)
                face_images.append(face_image)
            
            for num, face in enumerate(face_images, start=1):
                if vision:
                    cv2.imshow("Dlib Raw Image...", imutils.resize(raw_image, width=640))
                    cv2.imshow("Dlib Align faces...", imutils.resize(face, width=250))
                    cv2.waitKey(vision_millisecs)
                
                if save_dir != None:
                    if not os.path.exists(save_dir):
                        os.makedirs(save_dir)
                        logging.info("Builed {} dircetory successfully !".format(save_dir)) if os.path.exists(save_dir) else logging.warning("Builed {} directory failed !".format(save_dir))

                    image_path = os.path.join(save_dir, "{}.jpg".format(str(num).zfill(6)))
                    cv2.imwrite(image_path, cv2.resize(face, (face_size, face_size)), [cv2.IMWRITE_JPEG_QUALITY, 100])                    
                    logging.info("Saved image to '{}' successfully !".format(image_path)) if os.path.exists(image_path) else logging.warning("Saved image to '{}' failed !".format(image_path))
            cv2.destroyAllWindows()
        else:
            logging.warning("Dlib doesn't detect the face !")

        return (rois, raw_image, face_images)